<?php
namespace metagen_php;
use Exception;

function get_twig(array $inc_path = [])
{
  array_unshift($inc_path, __DIR__ . "/../tpl/");
  $loader = new \Twig\Loader\FilesystemLoader($inc_path);

  $twig = new \Twig\Environment($loader, [
    'debug' => true, 
    'autoescape' => false, 
    'strict_variables' => true]
  );
  $twig->addExtension(new \Twig\Extension\DebugExtension());

  _add_twig_support($twig);

  return $twig;
}

function supported_tokens()
{
  return [
    'default',
    'optional',
    'virtual',
    //TODO: anyone uses it?
    'is_enum_mask', 
  ];
}

function _add_twig_support(\Twig\Environment $twig)
{
  $twig->addTest(new \Twig\TwigTest('instanceof', 
    function($obj, $class) 
    { 
      return (new \ReflectionClass($class))->isInstance($obj);
    }
  ));
  $twig->addFilter(new \Twig\TwigFilter('var_export', 
    function($v) 
    { 
      return str_replace("\n", "", var_export($v, true));
    }
  ));
  $twig->addFunction(new \Twig\TwigFunction('Error', 
    function($e) 
    {
      throw new Exception($e);
    }
  ));
  $twig->addFunction(new \Twig\TwigFunction('has_token', 
    function($o, $token) 
    { 
      return $o->hasToken($token);
    }
  ));
  $twig->addFunction(new \Twig\TwigFunction('token', 
    function($o, $name) 
    {
      return $o->getToken($name);
    }
  ));
  $twig->addFilter(new \Twig\TwigFilter('default_value', 
    function($default) 
    { 
      return default_value($default);
    }
  ));
  $twig->addFunction(new \Twig\TwigFunction('apply_value_filters', 
    function ($field_name, array $tokens, $val, $add_assoc_check = true)
    {
      return apply_value_filters($field_name, $tokens, $val, $add_assoc_check);
    }
  ));
  $twig->addFunction(new \Twig\TwigFunction('data2value', 
    function($name, \mtgType $type, $buf, $prefix = '', $tokens = array(), $as_is = false, $postfix = '') 
    { 
      return data2value($name, $type, $buf, $prefix, $tokens, $as_is, $postfix);
    }
  ));
  $twig->addFunction(new \Twig\TwigFunction('value2data', 
    function($name, \mtgType $type, $buf, $prefix = '', $tokens = array()) 
    { 
      return value2data($name, $type, $buf, $prefix, $tokens);
    }
  ));
}

function default_value($default)
{
  if($default == "[]")
    return 'array()';

  if(strlen($default) > 2 && strpos($default, '[') === 0 && strpos($default, ']') === strlen($default)-1)
  {
    $str = trim($default, '][');
    if(strpos($str, "{") !== false )
      $str = var_export(json_decode(trim($default, ']['), true), true);
    return 'array(' . $str . ')';
  }
  return str_replace('$', '\$', $default);
}

function apply_filters($token_prefix, $field_name, array $tokens, $val, $add_assoc_check = true)
{
  $str = $val;
  foreach($tokens as $token => $args_json)
  {
    if(strpos($token, $token_prefix) === false)
      continue;

    $filter_func = '\metagen_php\\' . $token; 

    if(function_exists($filter_func))
    {
      $args = null;
      if($args_json && (strpos($args_json, '[') === 0 || strpos($args_json, '{') === 0))
      {
        $args = json_decode($args_json, true);
        if(!is_array($args))
          throw new Exception("Bad filter '$token' args: '$args_json'");
      }
      else
        $args = $args_json;

      if($add_assoc_check)
        $str = "(\$assoc ? $filter_func($str, '$field_name', \$this, " . str_replace("\n", "",  var_export($args, true)) . ") : $str)";
      else
        $str = "$filter_func($str, '$field_name', \$this, " . str_replace("\n", "",  var_export($args, true)) . ")";
    }
    else
      throw new Exception("No such filter '$filter_func'");
  }
  return $str;
}

function apply_array_filters($field_name, array $tokens, $val, $add_assoc_check = true)
{
  return apply_filters("fltarr_", $field_name, $tokens, $val, $add_assoc_check);
}

function apply_value_filters($field_name, array $tokens, $val, $add_assoc_check = true)
{
  return apply_filters("flt_", $field_name, $tokens, $val, $add_assoc_check);
}

//TODO: move it to template
function data2value($name, \mtgType $type, $buf, $prefix = '', $tokens = array(), $as_is = false, $postfix = '', $indent = '')
{
  $str = '';
  $tmp_val = '$tmp_val__';
  $pname = $prefix.$name;

  $cond_indent = $indent."  ";

  $default_value_arg = array_key_exists('default', $tokens) ? ', '.default_value($tokens['default']) : '';
  if($type instanceof \mtgMetaStruct)
  {
    $default = array_key_exists('default', $tokens) ? $tokens['default'] : null;
    if($default)
    {
      $default = json_decode($default, true);
      if(is_array($default))
      {
        $default = str_replace("\n", "",  var_export($default, true));
        $default_value_arg = ", \$assoc ? $default : array_values($default)";
      }
      else if($default === null)
      {
        $default_value_arg = ", null";
      }
      else
        throw new Exception("Bad default struct: " . $tokens['default']);
    }
  }

  $str .= "\n";

  if($as_is)
    $tmp_val = $buf;
  else
    $str .= $indent."{$tmp_val} = \metagen_php\array_extract_val({$buf}, \$assoc, '{$name}' {$default_value_arg});\n";

  if($type instanceof \mtgBuiltinType)
  {
    $str .= $cond_indent."{$tmp_val} = " . apply_value_filters($name, $tokens, "{$tmp_val}"). ";\n";
    $str .= $cond_indent."{$pname} = \metagen_php\\val_{$type}({$tmp_val});\n";
  }
  else if($type instanceof \mtgMetaStruct)
  {
    if(array_key_exists('virtual', $tokens))
    {
      $str .= $cond_indent."{$tmp_val} = " . apply_value_filters($name, $tokens, "{$tmp_val}"). ";\n";
      $str .= $cond_indent."\$tmp_sub_arr__ = \metagen_php\\val_arr({$tmp_val});\n";
      $str .= $cond_indent."\$vclass__ = AutogenBundle::getClassName(\metagen_php\\val_uint32(\metagen_php\array_extract_val(\$tmp_sub_arr__, \$assoc, 'vclass__')));\n";
      $str .= $cond_indent."{$pname} = new \$vclass__(\$tmp_sub_arr__, \$assoc);\n";
    }
    else
    {
      $str .= $cond_indent."if({$tmp_val} === null) {\n";
      $str .= $cond_indent."  {$pname} = null; \n";
      $str .= $cond_indent."} else {\n";
      $str .= $cond_indent."  {$tmp_val} = " . apply_value_filters($name, $tokens, "{$tmp_val}"). ";\n";
      $str .= $cond_indent."  \$tmp_sub_arr__ = \metagen_php\\val_arr({$tmp_val});\n";
      $str .= $cond_indent."  {$pname} = new {$type}(\$tmp_sub_arr__, \$assoc);\n";
      $str .= "}";
    }
  }
  else if($type instanceof \mtgArrType)
  {
    //TODO: maybe filters should be applied to the whole array as well?
    $str .= $cond_indent."{$tmp_val} = " . apply_array_filters($name, $tokens, "{$tmp_val}"). ";\n";
    $str .= $cond_indent."\$tmp_arr__ = \metagen_php\\val_arr({$tmp_val});\n";
    $str .= $cond_indent."foreach(\$tmp_arr__ as \$tmp_arr_item__)\n";
    $str .= $cond_indent."{\n";
    //NOTE: removing default for field
    unset($tokens['default']);
    $str .= data2value('$tmp__', $type->getValue(), "\$tmp_arr_item__", '', $tokens, true, "{$pname}[] = \$tmp__;", $cond_indent."  ") . "\n";
    $str .= $cond_indent."}\n";
  }
  else if($type instanceof \mtgMetaEnum)
  {
    $str .= $cond_indent."{$tmp_val} = " . apply_value_filters($name, $tokens, "{$tmp_val}"). ";\n";
    $check_enum_validity = array_key_exists('is_enum_mask', $tokens) ? 'false' : 'true';
    $str .= $cond_indent."{$pname} = \metagen_php\\val_enum('$type', {$tmp_val}, {$check_enum_validity});\n";
  }
  else
    throw new Exception("Unknown type '{$type}'");

  if($postfix)
    $str .= $cond_indent.$postfix."\n";

  //replacing some garbage
  $str = str_replace('$tmp_val__ = $tmp_val__;', '', $str);
  return $str;
}

//TODO: move it to template
function value2data($name, \mtgType $type, $buf, $prefix = '', $tokens = array(), $indent = '')
{
  $pname = $prefix.$name;

  $str = '';
  $str .= "\$__last_var = '$name';";
  $str .= "\$__last_val = $pname;";

  if($type instanceof \mtgBuiltinType)
  {
    if($type->isNumeric())
      $str .= $indent."\metagen_php\array_set_value({$buf}, \$assoc, '$name', 1*{$pname});";
    else if($type->isString())
      $str .= $indent."\metagen_php\array_set_value({$buf}, \$assoc, '$name', ''.{$pname});";
    else if($type->isBool())
      $str .= $indent."\metagen_php\array_set_value({$buf}, \$assoc, '$name', (bool){$pname});";
    else if($type->isBlob())
      $str .= $indent."\metagen_php\array_set_value({$buf}, \$assoc, '$name', {$pname});";
    else
      throw new Exception("Unknown type '$type'");
  }
  else if($type instanceof \mtgMetaStruct)
  {
    if(array_key_exists('virtual', $tokens))
      $str .= $indent."\metagen_php\array_set_value({$buf}, \$assoc, '$name', !is_object({$pname}) ? {$pname} : {$pname}->export(\$assoc, true/*virtual*/));";
    else
      $str .= $indent."\metagen_php\array_set_value({$buf}, \$assoc, '$name', !is_object({$pname}) ? {$pname} : {$pname}->export(\$assoc));";
  }
  else if($type instanceof \mtgMetaEnum)
  {
    $str .= $indent."\metagen_php\array_set_value({$buf}, \$assoc, '$name', 1*{$pname});";
  }
  else if($type instanceof \mtgArrType)
  {
    $str .= $indent."\$arr_tmp__ = array();\n";
    //NOTE: adding optimization, checking if the first item is array
    $str .= $indent."if(!\$assoc && {$pname} && is_array(current({$pname})))\n";
    $str .= $indent."{\n";
    //$str .= "  \metagen_php\debug('BULK:' . get_class(\$this));\n";
    $str .= $indent."  \$arr_tmp__ = {$pname};\n";
    $str .= $indent."}\n";
    $str .= $indent."else\n";
    $str .= $indent."  foreach({$pname} as \$idx__ => \$arr_tmp_item__)\n";
    $str .= $indent."  {\n";
    $str .= value2data('$arr_tmp_item__', $type->getValue(), "\$arr_tmp__", "", $tokens, $indent."    ") . "\n";
    $str .= $indent."    if(\$assoc)\n";
    $str .= $indent."    {\n";
    $str .= $indent."      \$arr_tmp__[] =  \$arr_tmp__['\$arr_tmp_item__'];\n";
    $str .= $indent."      unset(\$arr_tmp__['\$arr_tmp_item__']);\n";
    $str .= $indent."    }\n";
    $str .= $indent."  }\n";
    $str .= $indent."\metagen_php\array_set_value({$buf}, \$assoc, '$name', \$arr_tmp__);\n";
  }
  else
    throw new Exception("Unknown type '{$type}'");

  return $str;
}

function is_null_str($default)
{
  return is_string($default) && json_decode($default, true) === null;
}
